import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

def custom_boxplot(ax, data, mean, title, ylim, min_max=None, legend=False, x_labels=None):
    boxprops = dict(linestyle='-', linewidth=1., edgecolor='blue', facecolor='lightblue')
    whiskerprops = dict(color='black', linewidth=1.)
    capprops = dict(color='black', linewidth=1.)

    # make the median props as a solid black line
    medianprops = dict(color='black', linewidth=1.)

    # make the mean props as a dashed green line
    meanprops = dict(color='darkgreen', linestyle='--', linewidth=1.)

    if not isinstance(data, list):
        data = [data]
        mean = [mean]

    boxes = []
    outliers = []

    for i in range(len(data)):
        # extract outliers from the data (1 and 99 percentiles), i.e, first and last rows if min_max is None
        outliers.append(data[i].iloc[[0, -1]])

        # extract the box data (25th, 50th, 75th percentiles)
        box_data = data[i].iloc[2:5]

        # extract the whisker data (5th, 95th percentiles)
        whisker_data = data[i].iloc[[1, 6]]

        my_dict = {
            'med': box_data[1],
            'q1': box_data[0],
            'q3': box_data[2],
            'whislo': whisker_data[0],
            'whishi': whisker_data[1],
            'fliers': [],
            'mean': mean
        }

        boxes.append(my_dict)

    # Create boxplot manually
    ax.bxp(boxes, showmeans=False, meanline=True, widths=0.5,
        boxprops=boxprops, whiskerprops=whiskerprops,
        capprops=capprops, medianprops=medianprops,
        meanprops=meanprops, patch_artist=True)

    for i in range(len(mean)):
        label = "mean" if i == 0 else None
        ax.plot(i+1, mean[i], 'mx', label=label)

    # plot the outliers as white circles with a black border
    for i, outlier in enumerate(outliers):
        label = "1st, 99th" if i == 0 else None
        ax.plot([i+1, i+1], [outlier[0], outlier[1]], 'wo', markeredgecolor='black', label=label)

    if min_max is not None:

        if not isinstance(min_max, list):
            min_max = [min_max]

        for i, min_max_val in enumerate(min_max):
            label_min = "min" if i == 0 else None
            label_max = "max" if i == 0 else None
            # we plot the min and max values as well, min as a green triangle pointing down, max as a red triangle pointing up
            ax.plot(i+1, min_max_val[0], 'gv', label=label_min)
            ax.plot(i+1, min_max_val[1], 'r^', label=label_max)

            # check whether the min and max values are within the y limits
            # if not, we create a window within this plot to show them
            if min_max_val[0] < ylim[0]:
                # generate a rectangle to show the min value. put it right below the 1st percentile dot
                x_pos = 0.5*(2*i + 1)/len(data) - 0.05
                ax2 = ax.inset_axes([x_pos, 0.05, 0.1, 0.1])
                ax2.set_xticks([])
                # set the y-ticks to the min value and format it so that it has at most 2 decimal places
                min_2dp = float("{:.2f}".format(min_max_val[0]))
                ax2.set_yticks([min_2dp])

                ax2.plot(1, min_max_val[0], 'gv')
                ax2.set_xlim(0.5, 1.5)
                ax2.set_ylim(min_max_val[0] - min_max_val[0] * 0.1, min_max_val[0] + min_max_val[0] * 0.1)


            if min_max_val[1] > ylim[1]:
                # generate a rectangle to show the max value. put it right above the main plot
                x_pos = 0.5 * (2 * i + 1) / len(data) - 0.05
                ax3 = ax.inset_axes([x_pos, 0.85, 0.1, 0.1])
                ax3.set_xticks([])
                ax3.set_yticks([min_max_val[1]])
                ax3.plot(1, min_max_val[1], 'r^')
                ax3.set_xlim(0.5, 1.5)
                ax3.set_ylim(min_max_val[1] - min_max_val[1] * 0.1, min_max_val[1] + min_max_val[1] * 0.1)




    ax.set_title(title, fontsize=18, fontweight='bold', fontname='Arial')
    ax.set_ylim(ylim)
    # set y tick labels size
    ax.tick_params(axis='y', labelsize=15)
    # ax.set_xticklabels([y_label], fontsize=12)
    # remove xticks
    if x_labels is not None:
        ax.set_xticks([i + 1 for i in range(len(data))])
        ax.set_xticklabels(x_labels, fontsize=12)
    else:
        ax.set_xticks([])
    ax.grid(axis='y', linestyle='--', linewidth=0.5)

    if legend:
        # print a legend with font size 8
        ax.legend(loc="center right", fontsize=8)



if __name__ == "__main__":
    # Sample data
    data = {
        'Fidelity': [0.99, 0.98, 0.97, 0.96, 0.95],
        'Latency': [100, 150, 200, 250, 300] # milliseconds
    }

    """'Queue Size (non congested)': [0.1, 0.15, 0.2, 0.25, 0.3],  # num of requests
    'Queue Size (bottleneck link)': [10, 15, 20, 25, 30]  # num of requests"""

    csvs_lat = ["./out/latency_vector.csv", "./out2/latency_vector.csv"]
    csvs_fid = ["./out/fidelity_vector.csv", "./out2/fidelity_vector.csv"]
    x_labels = ["no RV", "with RV"]

    # replace the sample data with the data from the csv files
    df_list = []
    mean_fid_list = []
    min_max_fid_list = []
    mean_lat_list = []
    min_max_lat_list = []


    for i in range(len(csvs_lat)):
        df_latency = pd.read_csv(csvs_lat[i])
        # only keep the last half of the simulation
        df_latency = df_latency[df_latency["timestamp"] >= df_latency["timestamp"].max() / 2]
        df_latency["sample"] /= 1e3  # ms
        # compute boxplot data for latency, i.e., the 1st, 5th, 25th, 50th, 75th, 95th, 99th percentiles and the mean
        latency_boxplot_data = df_latency["sample"].describe(percentiles=[0.01, 0.05, 0.25, 0.5, 0.75, 0.95, 0.99])
        data['Latency'] = latency_boxplot_data[['1%', '5%', '25%', '50%', '75%', '95%', '99%']]

        df_fid = pd.read_csv(csvs_fid[i])
        # only keep the last half of the simulation
        df_fid = df_fid[df_fid["timestamp"] >= df_fid["timestamp"].max() / 2]
        # compute boxplot data for fidelity, i.e. the 1st, 5th, 25th, 50th, 75th, 95th, 99th percentiles and the mean
        fidelity_boxplot_data = df_fid["sample"].describe(percentiles=[0.01, 0.05, 0.25, 0.5, 0.75, 0.95, 0.99])
        data['Fidelity'] = fidelity_boxplot_data[['1%', '5%', '25%', '50%', '75%', '95%', '99%']]

        """df_queue_size_unc = pd.read_csv("./out/queue_size_free_vector.csv")
        # compute boxplot data for SKR, i.e. the 5th, 25th, 50th, 75th, 95th percentiles and the mean
        qs_unc_boxplot_data = df_queue_size_unc["sample"].describe(percentiles=[0.01, 0.05, 0.25, 0.5, 0.75, 0.95, 0.99])
        data['Queue Size (uncongested link)'] = qs_unc_boxplot_data[['1%', '5%', '25%', '50%', '75%', '95%', '99%']]
    
        df_queue = pd.read_csv("./out/queue_size_vector.csv")
        # compute boxplot data for throughput, i.e. the 5th, 25th, 50th, 75th, 95th percentiles and the mean
        queue_boxplot_data = df_queue["sample"].describe(percentiles=[0.01, 0.05, 0.25, 0.5, 0.75, 0.95, 0.99])
        data['Queue Size (bottleneck link)'] = queue_boxplot_data[['1%', '5%', '25%', '50%', '75%', '95%', '99%']]"""


        df = pd.DataFrame(data)
        df_list.append(df)

        mean_fid_list.append(df_fid["sample"].mean())
        min_max_fid_list.append((df_fid["sample"].min(), df_fid["sample"].max()))
        mean_lat_list.append(df_latency["sample"].mean())
        min_max_lat_list.append((df_latency["sample"].min(), df_latency["sample"].max()))

    # create custom boxplots from the percentiles we already have

    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(4, 4))

    # Fidelity
    # create a nice boxplot for the fidelity from scratch, without using the built-in boxplot function
    # we use the percentiles we already have
    # the boxplot consists of the 25th, 50th, and 75th percentiles, the whiskers are the 5th and 95th percentiles
    # and the outliers are the 1st and 99th percentiles
    # we also show the mean as a dashed line
    custom_boxplot(axes[0], [df["Fidelity"] for df in df_list], mean_fid_list, 'Fidelity', (0.5, 1), min_max_fid_list, True, x_labels)

    # add grids so that every y-tick has a grid line
    axes[0].grid(axis='y', linestyle='--', linewidth=0.5)

    # Latency
    # create a nice boxplot for the latency from scratch, without using the built-in boxplot function
    # we use the percentiles we already have
    # the boxplot consists of the 25th, 50th, and 75th percentiles, the whiskers are the 5th and 95th percentiles
    # and the outliers are the 1st and 99th percentiles
    # we also show the mean as a dashed line
    custom_boxplot(axes[1], [df["Latency"] for df in df_list], mean_lat_list, 'Latency (ms)', (0, 60), min_max_lat_list, False, x_labels)

    # add grids so that every y-tick has a grid line
    axes[1].grid(axis='y', linestyle='--', linewidth=0.5)


    """# Creating faceted plots
    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(6, 6))
    # fig.suptitle('Box Plots')

    # Fidelity
    df.boxplot(column=['Fidelity'], ax=axes[0, 0])
    axes[0, 0].set_title('Fidelity')
    print(df["Fidelity"])

    # Latency
    df.boxplot(column=['Latency'], ax=axes[0, 1])
    axes[0, 1].set_title('Latency (ms)')"""

    """# QSU
    df.boxplot(column=['Queue Size (uncongested link)'], ax=axes[1, 0])
    axes[1, 0].set_title('Queue Size (uncongested link)')

    # QSB
    df.boxplot(column=['Queue Size (bottleneck link)'], ax=axes[1, 1])
    axes[1, 1].set_title('Queue Size (bottleneck link)')"""

    # now we also highlight the sample mean
    """mean_latency = df_latency["sample"].mean()
    mean_fidelity = df_fid["sample"].mean()
    # mean_qsu = df_queue_size_unc["sample"].mean()
    # mean_queue = df_queue["sample"].mean()

    # plot the mean as a red dot in the boxplot
    axes[0, 0].plot(1, mean_fidelity, 'ro', label="sample mean")
    # axes[0, 0].legend(loc="lower right")
    axes[0, 0].set_ylim(0.5, 1)
    axes[0, 1].plot(1, mean_latency, 'ro', label="sample mean")
    axes[0, 1].set_ylim(0, 50)"""
    # axes[0, 1].legend(loc="upper right")
    """axes[1, 0].plot(1, mean_qsu, 'ro', label="sample mean")
    axes[1, 0].axhline(y=10, color='magenta', linestyle='--', label="reference queue size")
    axes[1, 0].set_ylim(0, 40)
    # axes[1, 0].legend(loc="upper right")
    axes[1, 1].plot(1, mean_queue, 'ro', label="sample mean")
    axes[1, 1].axhline(y=10, color='magenta', linestyle='--', label="reference queue size")
    axes[1, 1].set_ylim(0, 40)
    # axes[1, 1].legend(loc="upper right")"""

    plt.tight_layout() # Adjust the layout to make room for the main title

    # save the plot and make sure it is not cut off
    plt.savefig("./out/boxplots.pdf", bbox_inches='tight')

    plt.show()




from omnetpypy.front_end.port import Port


class SimulatedEntity:
    r"""
    A simulated entity is any active entity belonging to the simulation, such as modules and channels.
    Every entity may have a set of ports, and must keep a reference to the simulation context.
    
    Parameters
    ----------
    name : str
        The name of the entity. Must be unique within the simulation.
    identifier : int
        The unique identifier of the entity. Must be unique within the simulation.
    port_names : list of str
        The names of the ports of the entity.
    
    Attributes
    ----------
    name : str
        The name of the entity.
    identifier : int
        The unique identifier of the entity.
    sim_context : :class:`~omnetpypy.front_end.simulation.Simulation`
        The simulation where the entity is running. Used to access simulation variables like random number generators
        and the current simulation time.
    parent : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
        The parent entity of the current entity. If the entity is a top-level entity, the parent is None.
    is_listening : bool
        A flag indicating whether the entity is listening for incoming messages.
    ports : dict of :class:`~omnetpypy.front_end.port.Port`
        The ports of the entity, indexed by their names.
    """

    def __init__(self, name, identifier, port_names):
        self.name = name
        self.identifier = identifier
        self.sim_context = None
        self.parent = None

        self.is_listening = False
        # if the entity is listening, the connector will call the handle_message method when a message is received

        self.ports = {port_name: Port(port_name, parent=self) for port_name in port_names}

    def set_sim_context(self, sim_context):
        r"""
        Set the simulation context of the entity. Automatically called when the entity is added to the simulation.

        Parameters
        ----------
        sim_context : :class:`~omnetpypy.front_end.simulation.Simulation`
            The simulation context where the entity is running.
        """
        self.sim_context = sim_context

    def handle_message(self, message, port_name):
        r"""
        Process a message received as input to a port.

        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be processed.
        port_name : str or None
            The name of the port on which the message was received.
            If ``None``, the message was sent by the entity itself (self message).
        """
        pass

    def schedule_message(self, message, at=None, delay=None):
        r"""
        Schedule a self message to be received by this entity.
        Calls internally the method :meth:`~omnetpypy.front_end.simulation.Simulation.schedule_self_message`.
        
        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be processed.
        at : float or None, optional
            The simulation time at which the message should be processed.
            If ``None``, the ``delay`` parameter will be used.
        delay : float or None, optional
            The time delay from the current simulation time at which the message should be processed.
            If ``None``, the ``at`` parameter will be used.
        """
        # this sends a self message to the module at the specified time
        self.sim_context.connector.schedule_self_message(entity=self, message=message, at=at, delay=delay)

    def is_scheduled(self, message):
        r"""
        Check if a message is scheduled as a self message for this entity.
        Calls internally the method :meth:`~omnetpypy.front_end.simulation.Simulation.is_scheduled`.
        
        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be checked.
        """
        return self.sim_context.connector.is_scheduled(message, self)

    def send(self, message, port_name):
        r"""
        Send a message out on a specified port.
        Calls internally the method :meth:`~omnetpypy.front_end.port.Port.tx_output`.
        
        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be sent.
        port_name : str
            The name of the port on which the message should be sent.
        """
        self.ports[port_name].tx_output(message)

    def cancel_scheduled(self, message):
        r"""
        Cancel a scheduled self message for this entity.
        Calls internally the method :meth:`~omnetpypy.front_end.simulation.Simulation.cancel_scheduled`.
        
        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be cancelled.
        """
        self.sim_context.connector.cancel_scheduled(message, self)

    def initialize(self, step=0):
        r"""
        Initialize the entity right before the beginning of the simulation.
        This method is automatically called by the simulation context.

        The method is called 5 times, once for each step from 0 to 4. The step number is passed as a parameter.

        Parameters
        ----------
        step : int, optional
            The initialization step number. Default is 0. This parameter is used to allow entities to perform
            different initialization actions at different steps, and synchronize with other entities.
            The steps are numbered from 0 to 4, and the entity has the guarantee that the previous steps have
            already been executed on all entities of the simulation.
        """
        pass


"""
A set of utility classes and functions that are used across the package.
"""

import random
from numpy.random import default_rng

__all__ = ['MultiRandom', 'FutureMetric', 'get_metrics', 'get_metrics_from_csv', 'time_unit_factor']

from collections import namedtuple


class MultiRandom(random.Random):
    r"""
    This class is a wrapper around the standard library :class:`random.Random` class,
    that allows to use multiple random number generators with different seeds.

    Parameters
    ----------
    seeds : list or int
        A list of seeds for the random number generators.
        The length of the list determines the number of generators. If a single integer is provided,
        a single generator is created.
    """

    def __init__(self, seeds=44):
        if isinstance(seeds, int):
            seeds = [seeds]
        if len(seeds) == 0:
            super().__init__()
            self._generators = [self]
        else:
            self._generators = [random.Random() for _ in seeds]
            self.numpy_generators = [default_rng(seed=seed) for seed in seeds]
            for g, s in zip(self._generators, seeds):
                g.seed(s)

    def __len__(self):
        return len(self._generators)

    def random(self, generator=0):
        r"""
        Return the next random floating point number uniformly distributed in the range [0.0, 1.0).
        
        Parameters
        ----------
        generator : int, optional
            The index of the generator to use. Defaults to 0.
        
        Returns
        -------
        float
            The next random floating point number.
        
        See Also
        --------
        :meth:`random.Random.random`
        """
        return self._generators[generator].random()

    def randint(self, a, b, generator=0):
        r"""
        Return the next random integer N such that a <= N <= b.

        Parameters
        ----------
        a : int
            The lower bound of the random integer.
        b : int
            The upper bound of the random integer.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        int
            The next random integer.

        See Also
        --------
        :meth:`random.Random.randint`
        """
        return self._generators[generator].randint(a, b)

    def choice(self, seq, generator=0):
        r"""
        Return a random element from the non-empty sequence ``seq``.

        Parameters
        ----------
        seq : iterable
            A non-empty sequence.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        object
            A random element from the sequence.

        See Also
        --------
        :meth:`random.Random.choice`
        """
        return self._generators[generator].choice(seq)

    def choices(self, sequence, weights=None, cum_weights=None, k=1, generator=0):
        r"""
        Return a k sized list of elements chosen from the population with replacement.

        Parameters
        ----------
        sequence : iterable
            A non-empty sequence.
        weights : list, optional
            A list of weights to be used in the selection. If ``None``, the weights are assumed to be equal.
            The relative weights determine the probability of selecting each element.
        cum_weights : list, optional
            A list of cumulative weights to be used in the selection. If ``None``, the weights are assumed to be equal.
            The relative weights determine the probability of selecting each element.
        k : int, optional
            The number of elements to choose. Defaults to 1.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        list
            A list of k elements chosen from the population.

        See Also
        --------
        :meth:`random.Random.choices`
        """
        return self._generators[generator].choices(sequence, weights=weights, cum_weights=cum_weights, k=k)

    def shuffle(self, x, generator=0):
        r"""
        Shuffle the sequence x in place.

        Parameters
        ----------
        x : iterable
            The sequence to shuffle.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        See Also
        --------
        :meth:`random.Random.shuffle`
        """
        return self._generators[generator].shuffle(x)

    def sample(self, population, k, generator=0):
        r"""
        Return a k length list of unique elements chosen from the population sequence or set.

        Parameters
        ----------
        population : iterable
            A non-empty sequence or set.
        k : int
            The number of unique elements to choose.
        generator : int, optional
            The index of the generator to use. Defaults to 0.


        Returns
        -------
        list
            A list of unique elements chosen from the population.


        See Also
        --------
        :meth:`random.Random.sample`
        """
        return self._generators[generator].sample(population, k)

    def uniform(self, a, b, generator=0):
        r"""
        Return a random floating point number N such that a <= N <= b for a <= b and b <= N <= a for b < a.

        Parameters
        ----------
        a : float
            The lower bound of the random number.
        b : float
            The upper bound of the random number.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.


        See Also
        --------
        :meth:`random.Random.uniform`

        """
        return self._generators[generator].uniform(a, b)

    def geometric(self, p, size=None, generator=0):
        r"""
        Return a random integer N from a geometric distribution.

        This is the only method that uses the numpy random number generator (:class:`numpy.random.Generator`) instead
        of the standard library random number generator (:class:`random.Random`).

        Parameters
        ----------
        p : float
            The probability of success.
        size : int or tuple of ints or None, optional
            The shape of the output. If None, a single value is returned.
        generator : int, optional
            The index of the generator to use. Defaults to 0.


        See Also
        --------
        :meth:`numpy.random.Generator.geometric`
        """
        return self.numpy_generators[generator].geometric(p, size)

    def triangular(self, low=0.0, high=0.0, mode=None, generator=0):
        r"""
        Return a random floating point number N such that low <= N <= high and with the specified mode between bounds.

        Parameters
        ----------
        low : float
            The lower bound of the random number.
        high : float
            The upper bound of the random number.
        mode : float or None, optional
            The mode of the distribution. If None, the mode is the midpoint between the bounds.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        See Also
        --------
        :meth:`random.Random.triangular`
        
        """
        return self._generators[generator].triangular(low, high, mode)

    def normalvariate(self, mu=0.0, sigma=1.0, generator=0):
        r"""
        Return a random floating point number N from a normal (Gaussian) distribution.

        Parameters
        ----------
        mu : float
            The mean of the distribution.
        sigma : float
            The standard deviation of the distribution.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.normalvariate`
        """
        return self._generators[generator].normalvariate(mu, sigma)

    def lognormvariate(self, mu=0.0, sigma=1.0, generator=0):
        r"""
        Return a random floating point number N from a lognormal distribution.

        Parameters
        ----------
        mu : float
            The mean of the distribution.
        sigma : float
            The standard deviation of the distribution.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.lognormvariate`
        """
        return self._generators[generator].lognormvariate(mu, sigma)

    def gauss(self, mu=0.0, sigma=1.0, generator=0):
        r"""
        Return a random floating point number N from a normal (Gaussian) distribution.

        Parameters
        ----------
        mu : float
            The mean of the distribution.
        sigma : float
            The standard deviation of the distribution.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.gauss`
        """
        return self._generators[generator].gauss(mu, sigma)

    def expovariate(self, lambd=1.0, generator=0):
        r"""
        Return a random floating point number N from a normal (Gaussian) distribution.

        Parameters
        ----------
        lambd : float
            The rate of the exponential distribution.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.expovariate`
        """
        return self._generators[generator].expovariate(lambd)

    def vonmisesvariate(self, mu, kappa, generator=0):
        r"""
        Return a random floating point number N from a von Mises distribution.

        Parameters
        ----------
        mu : float
            The mean of the distribution.
        kappa : float
            The concentration of the distribution.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.vonmisesvariate`
        """
        return self._generators[generator].vonmisesvariate(mu, kappa)

    def gammavariate(self, alpha, beta, generator=0):
        r"""
        Return a random floating point number N from a gamma distribution.

        Parameters
        ----------
        alpha : float
            The shape of the distribution. Can be any positive number.
        beta : float
            The scale of the distribution. Can be any positive number.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.gammavariate`
        """
        return self._generators[generator].gammavariate(alpha, beta)

    def betavariate(self, alpha, beta, generator=0):
        r"""
        Return a random floating point number N from a beta distribution.

        Parameters
        ----------
        alpha : float
            The first shape parameter of the distribution. Can be any positive number.
        beta : float
            The second shape parameter of the distribution. Can be any positive number.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.betavariate`
        """
        return self._generators[generator].betavariate(alpha, beta)

    def paretovariate(self, alpha, generator=0):
        r"""
        Return a random floating point number N from a Pareto distribution.

        Parameters
        ----------
        alpha : float
            The shape of the distribution. Can be any positive number.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.parentovariate`
        """
        return self._generators[generator].paretovariate(alpha)

    def weibullvariate(self, alpha, beta, generator=0):
        r"""
        Return a random floating point number N from a Weibull distribution.

        Parameters
        ----------
        alpha : float
            The shape of the distribution. Can be any positive number.
        beta : float
            The scale of the distribution. Can be any positive number.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        Returns
        -------
        float
            The next random floating point number.

        See Also
        --------
        :meth:`random.Random.weibullvariate`
        """
        return self._generators[generator].weibullvariate(alpha, beta)

    def getstate(self, generator=0):
        r"""
        Return the internal state of the random number generator.

        Parameters
        ----------
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        See Also
        --------
        :meth:`random.Random.getstate`
        """
        return self._generators[generator].getstate()

    def setstate(self, state, generator=0):
        r"""
        Set the internal state of the random number generator.

        Parameters
        ----------
        state : tuple
            The internal state of the random number generator.
        generator : int, optional
            The index of the generator to use. Defaults to 0.

        See Also
        --------
        :meth:`random.Random.setstate`

        """
        self._generators[generator].setstate(state)


FutureMetric = namedtuple("FutureMetric", ["name", "vector", "mean", "median", "std", "var", "min", "max",
                               "count", "percentiles", "type", "columns"])
r"""
A type that describes the metrics to collect during simulations. The fields, apart from the name, are booleans
indicating the statistics to collect for the metric. Such fields are:

- vector: the complete list of samples of the metric 
- mean: the mean of the values 
- median: the median of the values 
- std: the standard deviation of the values 
- var: the variance of the values 
- min: the minimum value 
- max: the maximum value 
- count: the number of samples 
- percentiles: the percentiles of the values (1, 5, 25, 75, 95, 99) 

"""


def get_metrics(metric, df):
    r"""
    Compute the statistics for a metric and return them as a dictionary.
    Not used in the current version of the package because it requires all samples to be loaded in memory.

    Parameters
    ----------
    metric : :class:`~omnetpypy.utilities.FutureMetric`
        The metric to collect.
    df : :class:`pandas.DataFrame`
        The sampled data of the metric.

    Returns
    -------
    dict
        A dictionary containing the requested statistics for the metric.
    """

    # always use only the samples column and convert result to float

    final = {}
    if metric.vector:
        final["vector"] = df
    if metric.mean:
        final["mean"] = df['sample'].mean()
    if metric.median:
        final["median"] = df['sample'].median()
    if metric.std:
        final["std"] = df['sample'].std()
    if metric.var:
        final["var"] = df['sample'].var()
    if metric.min:
        final["min"] = df['sample'].min()
    if metric.max:
        final["max"] = df['sample'].max()
    if metric.count:
        final["count"] = df['sample'].count()
    if metric.percentiles:
        quantiles_df = df['sample'].quantile([0.01, 0.05, 0.25, 0.75, 0.95, 0.99])
        final["percentiles"] = quantiles_df.to_dict()

    return final


def get_metrics_from_csv(metric, filename):
    r"""
    Compute the statistics for a metric and return them as a dictionary.

    Parameters
    ----------
    metric : :class:`~omnetpypy.utilities.FutureMetric`
        The metric to collect.
    filename : string
        The name of the file containing the sampled data of the metric.

    Returns
    -------
    dict
        A dictionary containing the requested statistics for the metric.
    """

    # compute the statistics without loading the whole file in memory
    # the csv_iterator is an iterator over the rows of the csv file
    # we use the first row to get the column names
    # and we use the second row to initialize the statistics
    # then we iterate over the remaining rows to update the statistics

    import csv
    import numpy as np

    # read the first row to get the column names
    with open(filename, "r") as csv_iterator:
        csv_reader = csv.reader(csv_iterator)
        header = next(csv_reader)
        sample_index = header.index("sample")

        # use metric to initialize the statistics (except for the vector, which is ignored for the moment)
        final = {}
        samples = 0
        if metric.mean:
            final["mean"] = 0
        if metric.median:
            final["median"] = []
        if metric.std:
            final["std"] = 0
        if metric.var:
            final["var"] = 0
        if metric.min:
            final["min"] = np.inf
        if metric.max:
            final["max"] = -np.inf
        if metric.count:
            final["count"] = 0
        if metric.percentiles:
            final["percentiles"] = []

        # iterate over the remaining rows to update the statistics. Median and percentiles are computed online using
        # a list with 10K elements at most and a smaller list that collects the samples for the median as the median of
        # the 10k elements. This is done to avoid loading the whole file in memory.
        median_samples = []
        median_temp = []

        percentiles_samples = []
        percentiles_temp = []

        max_elems = 100000

        for row in csv_reader:
            samples += 1
            sample = float(row[sample_index])
            if metric.mean:
                final["mean"] += sample
            if metric.median:
                median_temp.append(sample)
                if len(median_temp) > max_elems - 1:
                    median_samples.append(np.median(median_temp))
                    median_temp = []
            if metric.std:
                final["std"] += sample**2
            if metric.var:
                final["var"] += sample**2
            if metric.min:
                final["min"] = min(final["min"], sample)
            if metric.max:
                final["max"] = max(final["max"], sample)
            if metric.count:
                final["count"] += 1
            if metric.percentiles:
                percentiles_temp.append(sample)
                if len(percentiles_temp) > max_elems - 1:
                    percentiles_samples.append(np.percentile(percentiles_temp, [1, 5, 25, 75, 95, 99]))
                    percentiles_temp = []

    if samples > 0:
        # compute the final statistics
        if metric.mean:
            final["mean"] /= samples
        if metric.median:
            if len(median_samples) == 0 or len(median_temp) > (max_elems // 100):
                median_samples.append(np.median(median_temp))
            final["median"] = np.mean(median_samples)
        if metric.std:
            final["std"] = np.sqrt(final["std"] / samples - final["mean"]**2)
        if metric.var:
            final["var"] = final["var"] / samples - final["mean"]**2
        if metric.count:
            final["count"] = samples
        if metric.percentiles:
            if len(percentiles_samples) == 0 or len(percentiles_temp) > (max_elems // 100):
                percentiles_samples.append(np.percentile(percentiles_temp, [1, 5, 25, 75, 95, 99]))
            # compute the final percentiles as a list of the sample mean of the respective percentiles in the samples
            final["percentiles"] = [np.mean([x[i] for x in percentiles_samples]) for i in range(6)]

    return final


def time_unit_factor(unit):
    r"""
    Return the factor to convert a time unit to seconds. For example, if the unit is "ms", the factor is 1e-3.

    Parameters
    ----------
    unit : str
        The time unit to convert. Can be one of the following:

            - "s" for seconds
            - "ms" for milliseconds
            - "us" for microseconds
            - "ns" for nanoseconds
    """

    if unit == "s":
        return 1
    elif unit == "ms":
        return 1e3
    elif unit == "us":
        return 1e6
    elif unit == "ns":
        return 1e9
    else:
        raise ValueError("Invalid time unit")


"""
We plot here fidelity and secret key rate in the same way as we did for latency and throughput in the previous notebook.
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

if __name__ == "__main__":
    fig, axs = plt.subplots(2, sharex=True)
    ax = axs[0]
    ax2 = axs[1]
    """
    ax.plot(df["timestamp"], df["sample"], label="congestion window")
    """

    ax2.set_xlabel("Time (ms)")
    # ax.set_ylabel("Congestion Window Size")
    ax.set_ylabel("Fidelity", color="red")
    ax.set_title("Fidelity and Secret Key Rate")

    # make the y-axis labels be at the same x position
    ax.yaxis.set_label_coords(x=-0.09, y=0.5)
    ax2.yaxis.set_label_coords(x=-0.09, y=0.5)

    # read also the file "latency_vector.csv" and plot a moving average of latency as a function of time in the same plot
    df_fid = pd.read_csv("./out/fidelity_vector.csv")

    # Compute the sliding window average
    window_size_time_units = 40000

    # create a new column with the sliding window index
    df_fid["sw_index"] = df_fid["timestamp"] // window_size_time_units

    # compute the sliding window average
    sw_avg = df_fid.groupby("sw_index").mean()

    # divide the timestamp by 1e3 to have ms
    sw_avg["timestamp"] /= 1e3

    ax.set_ylim(0.4, 1)

    # ax2 = ax.twinx()
    ax.plot(sw_avg["timestamp"], sw_avg["sample"], color="red", label="E2E Fidelity")
    # ax.set_ylabel("Latency (ms)")

    cur_flows = 4
    delete_phase = False
    # create a colormap for the background
    colors = {4: "blue", 10: "yellow", 16: "red"}
    for i in range(0, 48000, 8000):
        ax.axvline(x=i, color='black', linestyle=':', alpha=0.5)
        ax2.axvline(x=i, color='black', linestyle=':', alpha=0.5)
        # add some text next to the line saying "2 flows added" at the top
        ax.text(i + 1000, 0.5, f"{cur_flows} flows", rotation=0, fontsize=10)

        # add a colored background
        ax.axvspan(i, i + 8000, color=colors[cur_flows], alpha=0.1)
        ax2.axvspan(i, i + 8000, color=colors[cur_flows], alpha=0.1)

        if cur_flows == 16:
            delete_phase = True
        if cur_flows == 4:
            delete_phase = False
        if delete_phase:
            cur_flows -= 6
        else:
            cur_flows += 6

    # we obtain the secret key rate from df_fid by computing the secret key fraction for each sample
    # the secret key fraction is 1 - 2h(2(1-f)/3), where h is the binary entropy function and f is the fidelity
    df_skr = df_fid.copy()

    def binary_entropy(x):
        return -x * np.log2(x) - (1-x) * np.log2(1-x)

    df_skr["sample"] = 1 - 2 * binary_entropy(2 * (1 - df_skr["sample"]) / 3)
    # replace negative values with 0
    df_skr["sample"] = df_skr["sample"].clip(lower=0)

    # read the IRG file and plot the Inter-Request Gap (IRG) on the same plot but on a different axis
    window_size_time_units *= 5
    # sw_avg_irg = pd.Series([df_irg[(df_irg["timestamp"] >= t - window_size_time_units) & (df_irg["timestamp"] < t)]["sample"].mean() for t in df_irg["timestamp"]])
    sw_avg_skr = pd.Series(
        [df_skr[(df_skr["timestamp"] >= t - window_size_time_units) & (df_skr["timestamp"] < t)]["sample"].sum() for t
         in df_skr["timestamp"]])
    sw_avg_skr /= window_size_time_units  # pairs per us
    sw_avg_skr *= 1e3  # pairs per ms

    df_skr["timestamp"] /= 1e3

    ax2.plot(df_skr["timestamp"], sw_avg_skr, color="blue", label="E2E Secret Key Rate")
    # ax2.set_ylabel("Inter-Request Gap (IRG) (ms)", color="blue")
    ax2.set_ylabel("E2E Secret Key Rate (bits/ms)", color="blue")
    # ax2.set_ylim(0, 2)

    # print the legends within the axis

    lines, labels = ax.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    # lines3, labels3 = ax3.get_legend_handles_labels()
    # ax2.legend(lines + lines2, labels + labels2, loc="upper right")

    # print legend in the axis box
    # ax.legend(lines, labels, loc="best")
    # ax2.legend(lines2, labels2, loc="best")

    # adjust plot size
    fig.tight_layout()

    plt.show()

    # save the figure
    fig.savefig("./out/fidelity_secret_key_rate.pdf")


"""
This module implements the SimpleModule abstract class.

The class has the same semantic as its homonym in omnet++, and it is meant to be subclassed by the user to define
its custom simulation modules.
"""
from abc import abstractmethod

from omnetpypy.front_end.sim_entity import SimulatedEntity


class SimpleModule(SimulatedEntity):
    r"""
    This class is an abstract class that represents a simple module in the simulation.
    Simple modules are the basic building blocks of the simulation model. They can send and receive messages
    through their ports, and they can be connected to other modules through channels connecting their ports.
    The behavior of a simple module is defined by how it handles incoming messages. The user should subclass this class
    to define custom simulation modules.

    Simple modules are also in charge of recording metrics samples. The user can call the method
    :meth:`~omnetpypy.front_ent.simple_module.SimpleModule.emit_metric` at any time to record a metric sample.

    See :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity` for inherited attributes.

    Parameters
    ----------
    name : str
        The name of the module. This name should be unique within the simulation.
    identifier : int
        The identifier of the module. This identifier should be unique within the simulation.
    port_names : list of str
        The names of the ports of the module.
    """

    def __init__(self, name, identifier, port_names):
        super().__init__(name, identifier, port_names)
        self.is_listening = True

    @abstractmethod
    def handle_message(self, message, port_name):
        r"""
        Handle a message received as input to a port.
        This method must be implemented by every subclass to define the behavior of the custom module.

        Parameters
        ----------
        message : :class:`~omnetpypy:front_end.message.Message`
            The message to be processed.
        port_name : str or None
            The name of the port on which the message was received.
            If ``None``, the message is a self message scheduled by this module.
        """
        pass

    def connect(self, local_port, remote_entity, remote_port, channel=None):
        r"""
        Connect a port of this module to a port of another remote entity.
        The output of a port will be fed as input to the other port
        (up to the intermediate actions of a channel, if any).

        Parameters
        ----------
        local_port : str
            The name of the local port to be connected.
        remote_entity : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
            The remote entity to which the port will be connected.
        remote_port : str
            The name of the remote port to be connected.
        channel : :class:`~omnetpypy.front_end.channel.Channel` or None, optional
            The channel through which the connection will be made, if any.
            If set, the local port is connected to the "A" port of the channel,
            and the "B" port of the channel is connected to the remote port.
        """

        if channel is not None:
            self.ports[local_port].connect(channel.ports["A"])
            channel.ports["B"].connect(remote_entity.ports[remote_port])
        else:
            self.ports[local_port].connect(remote_entity.ports[remote_port])

    def emit_metric(self, name, value):
        r"""
        Record a metric sample in the simulation context.
        The metric name must be defined in the main configuration file.

        Parameters
        ----------
        name : str
            The name of the metric.
            The metric name must be defined in the simulation configuration.
        value : Any
            The value of the metric sample
        """
        self.sim_context.connector.record_metric(name, value)



"""
This module implements the :meth:`~omnetpypy.front_end.compound_module.CompoundModule` class.
Compound modules are non-atomic modules that contain other modules and channels.
They do not come with a behavior, but they can be used to organize the simulation model with a hierarchical structure.
Their behavior is intrinsically defined by the behavior of their submodules and how they are connected.

Compund Modules have the same semantic as in omnet++
"""
from omnetpypy.front_end.sim_entity import SimulatedEntity


class CompoundModule(SimulatedEntity):
    r"""
    This class represents a compound module in the simulation model.
    Compound modules are non-atomic modules that contain other modules and channels.
    They do not come with a behavior, but they can be used to organize the simulation model with a hierarchical
    structure. Their behavior is intrinsically defined by the behavior of their submodules and how they are connected.

    See :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity` for inherited attributes.
    All kwargs passed to the constructor are also stored as attributes of the module.

    Parameters
    ----------
    name : str
        The name of the module. This name should be unique within the simulation.
    identifier : int
        The identifier of the module. This identifier should be unique within the simulation.
    port_names : list of str
        The names of the ports of the module.
    **kwargs
        Additional, arbitrary attributes of the module, passed as keyword arguments.

    Attributes
    ----------
    sub_modules : dict of :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
        The sub-entities of the module, indexed by their names.
    """

    def __init__(self, name, identifier, port_names, **kwargs):
        super().__init__(name, identifier, port_names)

        # additional attributes based on kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.sub_modules = {}

    def add_sub_module(self, module):
        r"""
        Add a submodule to this compound module.

        Parameters
        ----------
        module : :class:`~omnetpypy:front_end.sim_entity.SimulatedEntity`
            The submodule to be added.

        Raises
        ------
        Exception
            If a submodule with the same name already exists.
        """
        if module.name in self.sub_modules:
            raise Exception("A submodule with the same name already exists")
        self.sub_modules[module.name] = module
        module.parent = self

    def get_sub_module(self, name):
        r"""
        Get a submodule by its name.

        Parameters
        ----------
        name : str
            The name of the submodule to be retrieved.

        Returns
        -------
        :class:`~omnetpypy:front_end.sim_entity.SimulatedEntity`
            The submodule with the given name.
        """
        return self.sub_modules[name]

    def forward_input(self, port_container, port_submodule):
        r"""
        Forward the input of a port of this module to a port of a submodule.

        Parameters
        ----------
        port_container : str
            The name of the port of this module that will forward the input.
        port_submodule : :class:`~omnetpypy:front_end.port.Port`
            The port of the submodule that will receive the input.

        See Also
        --------
        :meth:`~omnetpypy.front_end.port.Port.forward_input`
        """
        self.ports[port_container].forward_input(port_submodule)

    def forward_output(self, port_submodule, port_container):
        r"""
        Forward the output of a port of a submodule to a port of this module.

        Parameters
        ----------
        port_submodule : str
            The name of the port of the submodule that will forward the output.
        port_container : :class:`~omnetpypy:front_end.port.Port`
            The port of this module that will receive the output.

        See Also
        --------
        :meth:`~omnetpypy.front_end.port.Port.forward_output`
        """
        self.ports[port_submodule].forward_output(port_container)

    def connect(self, local_port, remote_entity, remote_port, channel=None):
        r"""
        Connect a port of this module to a port of another remote entity.

        Parameters
        ----------
        local_port : str
            The name of the local port to be connected.
        remote_entity : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
            The remote entity to which the port will be connected.
        remote_port : str
            The name of the remote port to be connected.
        channel : :class:`~omnetpypy.front_end.channel.Channel` or None, optional
            The channel through which the connection will be made, if any.
            If set, the local port is connected to the "A" port of the channel,
            and the "B" port of the channel is connected to the remote port.

        See Also
        --------
        :meth:`~omnetpypy.front_end.port.Port.connect`
        """

        if channel is not None:
            self.ports[local_port].connect(channel.ports["A"])
            channel.ports["B"].connect(remote_entity.ports[remote_port])
        else:
            self.ports[local_port].connect(remote_entity.ports[remote_port])

    def schedule_message(self, message, at=None, delay=None):
        raise Exception("CompoundModule does not support scheduling messages")

    def handle_message(self, message, port_name):
        raise Exception("CompoundModule does not support handling messages")

    def send(self, message, port_name):
        raise Exception("CompoundModule does not support sending messages")

    def cancel_scheduled(self, message):
        raise Exception("CompoundModule does not support cancelling messages")

    def is_scheduled(self, message):
        raise Exception("CompoundModule does not support checking if a message is scheduled")


r"""
This module implements the :meth:`~omnetpypy.front_end.channel.Channel` class,
which implements advanced connectivity between ports in the simulation.
"""
from omnetpypy.front_end import Message
from omnetpypy.front_end.sim_entity import SimulatedEntity


class Channel(SimulatedEntity):
    r"""
    Channels are used to apply operations on messages travelling between connected ports.
    They can apply noise, delays, losses, and other custom operations to the messages.

    Channels always have two ports, named "A" and "B". The channel forwards messages
    from port "A" to port "B", and vice versa, after applying the specified delay and loss probability.

    Subclasses can override the methods :meth:`~omnetpypy.front_end.channel.Channel.process_message` to apply more
    complex and asymmetric operations on the messages passing through the channel.

    Parameters
    ----------
    name : str
        The name of the channel. This name should be unique within the simulation.
    identifier : int or None, optional
        The identifier of the channel. This identifier should be unique within the simulation.
        If ``None``, the identifier will be automatically generated.
    delay : float or None, optional
        The delay to be applied to the messages passing through the channel. If ``None``, no delay is applied.
    loss_prob : float or None, optional
        The probability of loss to be applied to the messages passing through the channel.
        If ``None``, no loss is applied.

    Attributes
    ----------
    delay : float or None
        The delay to be applied to the messages passing through the channel.
        If ``None``, no delay is applied.
    loss_prob : float or None
        The probability of loss to be applied to the messages passing through the channel.
        If ``None``, no loss is applied.
    """

    def __init__(self, name, identifier=None, delay=None, loss_prob=None):
        super().__init__(name, identifier, port_names=["A", "B"])
        self.is_listening = True
        self.delay = delay
        self.loss_prob = loss_prob

    def handle_message(self, message, port_name):
        r"""
        Handle a message received from a port. First, it processes the message  by calling the method
        :meth:`~omnetpypy.front_end.channel.Channel.process_message`, then it applies the
        optional delay and loss probability, and finally it sends the message out of the other port (if not lost).

        Parameters
        ----------
        message : Message
            The message to be handled.
        port_name : str
            The port from which the message was received.
        """

        if port_name is None:  # this is a self message, should be immediately sent to the other port
            port = message.meta["port_noneshouldusethiskey"]
            del message.meta["port_noneshouldusethiskey"]
            self.send(message, port)
            return

        post_processed_message = self.process_message(message, port_name)

        should_drop = self.apply_loss(message, port_name) or post_processed_message is None
        out_port = "B" if port_name == "A" else "A"
        if not should_drop and (self.delay is None or self.delay <= 0):
            self.send(post_processed_message, out_port)
        elif not should_drop:
            post_processed_message.meta["port_noneshouldusethiskey"] = out_port
            self.schedule_message(post_processed_message, delay=self.generate_delay(message, port_name))

    def process_message(self, message, port_name):
        r"""
        Process a message received from a port. By default, this method does nothing and returns the message as is.
        Subclasses can override this method to apply more complex operations on the messages passing through the
        channel.

        Parameters
        ----------
        message : :meth:`~omnetpypy.front_end.message.Message`
            The message to be processed.
        port_name : str
            The port from which the message was received.

        Returns
        -------
        :meth:`~omnetpypy.front_end.message.Message` or None
            The message to be sent to the connected port, or ``None`` if the message should be dropped.
        """

        # default behavior
        return message

    def generate_delay(self, message, port_name):
        """
        Generate a delay for the message, based on an arbitrary delay distribution.

        Parameters
        ----------
        message : Message
            The message to be delayed.
        port_name : str
            The port from which the message was received.

        Returns
        -------
        float or None
            The delay to be applied to the message. If None or zero, the message will be sent immediately.
        """

        # default behavior
        return self.delay

    def apply_loss(self, message, port_name):
        """
        Apply loss to the message, based on an arbitrary loss probability distribution.

        Parameters
        ----------
        message : Message
            The message to be delayed.
        port_name : str
            The port from which the message was received.

        Returns
        -------
        bool
            True if the message should be lost, False otherwise.
        """

        # default behavior
        if self.loss_prob is not None:

            # we use the default rng in the sim_context
            return self.sim_context.rng.random() < self.loss_prob

        return False

